#include <cmath>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <boost/dynamic_bitset.hpp>
#include "Isomorphism.h"

using namespace std;
using namespace boost;

//#define DEBUG

Isomorphism::Isomorphism(vector<string> pattern_) {
	vector<string> groups = pattern_;
	pattern = pattern_;
	int n_elements = pattern.size();

	// get unique groups of elements in the module: use string unique operation
	vector<string>::iterator it;
	sort(groups.begin(), groups.end());
	it = unique (groups.begin(), groups.end());
	groups.resize( distance(groups.begin(),it) );
	int n_groups = groups.size();

	// get "group2loc", where are the elements of each group located?
	for (int i=0; i<n_groups; i++) {
		string& g = groups[i];
		for (int j=0; j<n_elements; j++) {
			if (pattern[j].compare(g)==0) {
				group2loc[g].push_back(j);
			}
		}
	}

	// get bit_set for each possible flipping of groups.
	ISOMORPHISM_CODING_NUMBER_TYPE n_possible_flips = (ISOMORPHISM_CODING_NUMBER_TYPE)pow(2.0,n_groups) - 1; // Each group can have a label "A" or "B". So we have "2^n_groups" possible labels
	vector<dynamic_bitset<> > bits_for_possible_group_labels(n_possible_flips); // all 0's by default
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++) { // We don't need all bits ZERO. All bits ZERO means it doesn't flip.
		bits_for_possible_group_labels[i] = dynamic_bitset<>(n_groups, i+1); // assign i to this bit_set
	}
#ifdef DEBUG
	cout << "pattern: " << endl;
	for (int i=0; i<pattern.size(); i++)
		cout << pattern[i] << ", ";
	cout << endl;
	map<string,vector<int> >::iterator iter;
	for (iter=group2loc.begin(); iter!=group2loc.end(); iter++) {
		cout << iter->first << ":\t";
		for (int i=0; i<iter->second.size(); i++)
			cout << iter->second[i] << ", ";
		cout << endl;
	}
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++)
		cout << "bits(" << i << ") = " << bits_for_possible_group_labels[i] << ", " << bits_for_possible_group_labels[i].to_ulong() << endl;
#endif

	// expand group flipping to elements flipping
	// element flipping: we flip labels of all the elements in the same group.
	bits_for_flip_elements.reserve(n_possible_flips);
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++) {
		bits_for_flip_elements.push_back(dynamic_bitset<>(n_elements, 0)); // all 0's by default
		for (int j=0; j<n_groups; j++) {
			string& g = groups[j]; // group ID
			bool flip_or_not = bits_for_possible_group_labels[i][j]; // 1 is to flip, 0 is not to flip.
			for (int k=0; k<group2loc[g].size(); k++) { // visit all elements of this group
				int loc = group2loc[g][k];
				bits_for_flip_elements[i][loc] = flip_or_not;
			}
		}
	}
#ifdef DEBUG
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++) {
		cout << "flip_bits(" << i << ") = " << bits_for_flip_elements[i] << ", " << bits_for_flip_elements[i].to_ulong() << endl;
	}
#endif
}

void Isomorphism::getIsomorphisms(ISOMORPHISM_CODING_NUMBER_TYPE module, vector<ISOMORPHISM_CODING_NUMBER_TYPE>& iso_modules) {
	iso_modules.clear();
	int n_elements = pattern.size();
	int n_possible_flips = bits_for_flip_elements.size();
	dynamic_bitset<> x(n_elements, module);
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++) {
		dynamic_bitset<> y = x ^ bits_for_flip_elements[i];  // xor operation: flip the label when bit is 1, otherwise remain unchanged when bit is 0.
		iso_modules.push_back(y.to_ulong());
	}
	sort(iso_modules.begin(), iso_modules.end()); // sort it in increasing order
#ifdef DEBUG
	cout << "query_module = " << x << ", " << x.to_ulong() << endl;
	for (ISOMORPHISM_CODING_NUMBER_TYPE i=0; i<n_possible_flips; i++) {
		dynamic_bitset<> y(n_elements, iso_modules[i]);
		cout << "isomorphisms(" << i << ") = " << y << ", " << iso_modules[i] << endl;
	}
#endif
}

void Isomorphism::getPatternOfIsomorphism(ISOMORPHISM_CODING_NUMBER_TYPE module,
	vector<string>& prefix, vector<string>& result_pattern) {
	int n_elements = pattern.size();
	dynamic_bitset<> x(n_elements, module);
	for (int i=0; i<n_elements; i++) {
		if (x[i])
			result_pattern.push_back(prefix[i] + ".B");
		else
			result_pattern.push_back(prefix[i] + ".A");
	}
}

void Isomorphism::getPatternOfIsomorphism(ISOMORPHISM_CODING_NUMBER_TYPE module,
	map<int,vector<int> >& chromregion2bead, vector<int>& chrom_regions_module,
	vector<int>& result_pattern) {
	int n_elements = pattern.size();
	dynamic_bitset<> x(n_elements, module);
	for (int i=0; i<n_elements; i++) {
		if (x[i]) // x[i] is 0 or 1.
			result_pattern.push_back(chromregion2bead[chrom_regions_module[i]][1]);
		else
			result_pattern.push_back(chromregion2bead[chrom_regions_module[i]][0]);
	}
}
