#include <sstream>
#include <iostream>
#include <fstream>
#include <map>
#include <vector>
#include <string>
#include "utils.h"
#include "Modules.h"

using namespace std;

Module::Module(const char* filename, int which_module)
{
	module_index = which_module;
	read_module_from_file(filename, which_module, nodes, networks);
}

// Input:
//   In the modules file, each line contains a module's info.
//   "which_module" is a 1-based index (or line number) of a module.
// Output:
//   "nodes" is a list of 0-based index of nodes in the module.
//   "networks" is a list of 0-based index of networks where this module occurs.
void Module::read_module_from_file(const char* filename, int which_module, vector<int>& nodes, vector<int>& networks)
{
	cout << "read_module_from_file()" << endl;
	cout << "\tfile: " << filename << endl;
	cout << "\t#module_index: " << which_module << endl;
	nodes.clear();
	networks.clear();
	ifstream modin;
	modin.open(filename);
	if (modin.fail()){
        cerr << "Error: Unable to open " << filename << " in read_module_from_file()" << endl;
        exit(EXIT_FAILURE);
    }
	string buffer;
	// skip the first (which_module-1) lines or modules.
	int i;
	for (i=1; i<which_module && !modin.eof(); i++)
		getline(modin, buffer);
	if (i<which_module) {
		cerr << "Error: only " << i-1 << " modules in the file, which is < module_index(1-based)=" << which_module << endl;
		exit(EXIT_FAILURE);
	}
	getline(modin, buffer); // this line contains the module we would like to have.
	modin.close();

	vector <string> fields;
	split(buffer, "\t", fields);

	// get list of nodes
	string s = fields[0].substr(1, fields[0].size()-2);
	split_int(s, ", ", nodes); // the 1st item contains networks info. remove "[" and "]" from the beginning and end of this string
	for (int i=0; i<nodes.size(); i++)
		nodes[i]--;   // node index is 0-based.
	cout << "\t#nodes_module: " << nodes.size() << endl;
//	print( nodes, ", " );

	// get list of networks
	vector<string> sub_fields;
	split(fields[6], " ", sub_fields); // the 7th item contains networks info.
	for (int i=0; i<sub_fields.size(); i++) {
		vector<string> items;
		split(sub_fields[i], ":", items);
		istringstream iss(items[1]);
		int v;
		iss >> v;
		networks.push_back( v-1 );  // network index is 0-based.
	}
	cout << "\t#networks_module: " << networks.size() << endl;
//	print( networks, ", " );
}
