#include <sstream>
#include <fstream>
#include <cassert>
#include <algorithm> // std::binary_search, std::sort
#include <boost/algorithm/string.hpp> // boost:trim
#include "Networks.h"

using namespace std;
using namespace boost::algorithm;

//#define DEBUG

ostream& operator<<(std::ostream& os, const Edge& e)
{
	return os << e.get_node_A()+1 << "\t" << e.get_node_B()+1;
}

Edge& Edges::operator [](int index)
{
	if (index>=0 && index<edges.size()) {
		return edges[index];
	} else {
		cerr << "Error: edge index must be 0<= <#edge!" << endl;
		exit(EXIT_FAILURE);
	}
}

ostream& operator<<(ostream& os, Edges& es)
{
	for (int i=0; i<es.size(); i++) {
		os << es[i] << endl;
	}
    return os;
}

Network::Network(string net_file, const char *file_type,
	vector<int> &selected_nodes_idx_set)
{
	if (strcmp(file_type,MULE_GRAPH_FORMAT)==0) {
		read_network_muleGraphFormat( net_file.c_str(), selected_nodes_idx_set, es );
	}
}

void Network::print(string out_file)
{
	ofstream out;
	cout << "#node_actual: " << get_actual_node_num() << endl;
	cout << "#edge_actual: " << get_actual_edge_num() << endl;
//	if (es!=NULL) {
	out.open(out_file.c_str());
		out << es;
		out.close();
	//} else {
	//	cerr << "Warning: Network is empty. No output!" << endl;
	//}
}

double Network::density()
{
	int n_actual_node=get_actual_node_num();
	int n_actual_edge=get_actual_edge_num();
	return (double)n_actual_edge*2/(n_actual_node*(n_actual_node-1));
}

// nodes are indexed for the whole network, not for this subgraph.
double Network::density_subgraph(vector<int>& subgraph_nodes_idx_set)
{
	int n_subgraph_nodes = subgraph_nodes_idx_set.size();
	int n_subgraph_edges = 0;
	vector<int> sorted_nodes(subgraph_nodes_idx_set);
	sort(sorted_nodes.begin(), sorted_nodes.end());
	for (int i=0; i<es.size(); i++) {
		Edge& e = es[i];
		int idxNode1 = e.get_node_A();
		bool exist_node1 = binary_search(sorted_nodes.begin(), sorted_nodes.end(), idxNode1);
		if (!exist_node1) continue;
		int idxNode2 = e.get_node_B();
		bool exist_node2 = binary_search(sorted_nodes.begin(), sorted_nodes.end(), idxNode2);
		if (exist_node2) {
			n_subgraph_edges++;
#ifdef DEBUG
			cout << idxNode1+1 << "\t" << idxNode2+1 << endl;
#endif
		}
	}
	double density = (double)n_subgraph_edges*2/(n_subgraph_nodes*(n_subgraph_nodes-1));
#ifdef DEBUG
	cout << "density_subgraph(#node=" << n_subgraph_nodes << ",#edge=" << n_subgraph_edges << ") = " << density << endl;
#endif
	return density;
}

/*
mule graph format:
  A graph is specified in a text file as follows:
  - The first line contains some identifier for the graph
  - The second line contains two integers, first being the number of
    nodes in the graph, second being the number of edges.
  - Each subsequent two lines correspond to a single node of the graph,
    where
    * The first line contains some identifier for this node
    * The second line contains the indexes of all nodes that share an
      edge for this node. If the graph is undirected, the edge needs to
      be listed only once based on some ordering of node identifiers.
      Note that indexing of nodes starts from 0.

"selected_nodes_idx_set" is a set of nodes index (starting from 0) which form a subgraph,
we return this subgraph as a network.
*/
void Network::read_network_muleGraphFormat(const char* file, vector<int>& selected_nodes_idx_set,
	Edges& es)
{
	es.clear();
	n_nodes_involved = selected_nodes_idx_set.size();
	vector<int> sorted_nodes(selected_nodes_idx_set);
	sort(sorted_nodes.begin(), sorted_nodes.end());
	ifstream gin;
	gin.open(file);
	if (gin.fail()){
        cerr << "Error: Unable to open " << file << " in read_network_muleGraphFormat()" << endl;
        exit(EXIT_FAILURE);
    }
	string buffer;
	getline(gin, buffer); // 1st line is this graph name
	getline(gin, buffer); // 2nd line is #gene and #edge
	istringstream iss(buffer);
	iss >> n_node >> n_edge;
//	cout << n_node << "\t" << n_edge << endl;
	for (int idxNode1=0; idxNode1<n_node; idxNode1++) {
		getline(gin, buffer); // this line is node label, skip it.
		getline(gin, buffer); // this line is a list of indexes of neighoured nodes. index starts from 0.
		bool exist_node1 = binary_search(sorted_nodes.begin(), sorted_nodes.end(), idxNode1);
		if (!exist_node1) {
			continue;
		}
		trim(buffer);
		if (buffer.empty()) continue;
		istringstream iss(buffer);
		while (iss.good()) {
			int idxNode2;
			iss >> idxNode2;
			bool exist_node2 = binary_search(sorted_nodes.begin(), sorted_nodes.end(), idxNode2);
			if (exist_node2)
				es.push_back(idxNode1, idxNode2);
		}
	}
	gin.close();
}

/*
mule graph format:
  A graph is specified in a text file as follows:
  - The first line contains some identifier for the graph
  - The second line contains two integers, first being the number of
    nodes in the graph, second being the number of edges.
  - Each subsequent two lines correspond to a single node of the graph,
    where
    * The first line contains some identifier for this node
    * The second line contains the indexes of all nodes that share an
      edge for this node. If the graph is undirected, the edge needs to
      be listed only once based on some ordering of node identifiers.
      Note that indexing of nodes starts from 0.

"selected_nodes_idx_set" is a set of nodes index (starting from 0) which form a subgraph,
we return this subgraph as a network.
*/
void Network::read_network_muleGraphFormat(const char* file, Edges& es)
{
	es.clear();
	ifstream gin;
	gin.open(file);
	if (gin.fail()){
        cerr << "Error: Unable to open " << file << " in read_network_muleGraphFormat()" << endl;
        exit(EXIT_FAILURE);
    }
	string buffer;
	getline(gin, buffer); // 1st line is this graph name
	getline(gin, buffer); // 2nd line is #gene and #edge
	istringstream iss(buffer);
	iss >> n_node >> n_edge;
//	cout << n_node << "\t" << n_edge << endl;
	for (int idxNode1=0; idxNode1<n_node; idxNode1++) {
        getline(gin, buffer); // this line is node label, skip it.
		getline(gin, buffer); // this line is a list of indexes of neighoured nodes. index starts from 0.
		trim(buffer);
		if (buffer.empty()) continue;
		istringstream iss(buffer);
		while (iss.good()) {
			int idxNode2;
			iss >> idxNode2;
			es.push_back(idxNode1, idxNode2);
		}
	}
	gin.close();
	if (n_edge != es.size()) {
		cout << "\nWarning: " << file << " #supposed_edge=" << n_edge <<", #actual_edge=" << es.size();
		cerr << "\nWarning: " << file << " #supposed_edge=" << n_edge <<", #actual_edge=" << es.size();
	}
}

Networks::Networks(const char* nets_path, vector<string>& nets_IDs_list, const char* net_file_suffix,
		const char* file_type, vector<int>& selected_nodes_idx_set)
{
}