#ifndef NETWORKS_H
#define NETWORKS_H

#include <sstream>
#include <iostream>
#include <map>
#include <vector>
#include <string>

#define MULE_GRAPH_FORMAT "MULE"
#define EDGE_LIST_GRAPH_FORMAT "EDGE_LIST"

class Edge {
private:
	int node_A;
	int node_B;
public:
	Edge(int A, int B) {node_A=A; node_B=B;};
	int get_node_A() const {return node_A;};
	int get_node_B() const {return node_B;};
	friend std::ostream& operator<<(std::ostream& os, const Edge& e);
};

class Edges {
private:
	std::vector<Edge> edges;
public:
	Edges() {};
	void push_back(int A, int B) {push_back(Edge(A,B));};
	void push_back(const Edge& e) {edges.push_back(e);};
	void clear() {edges.clear();};
	int size() {return edges.size();};
	Edge& operator [](int index);
	friend std::ostream& operator<<(std::ostream& os, Edges&);
};

class Network {
private:
	int n_node; // #nodes in the original network
	int n_edge; // #edges in the original network
	Edges es; // actual edges involved
	int n_nodes_involved; // actual nodes involved
public:
	Network() {};
	Network(std::string net_file, const char *file_type, std::vector<int>& selected_nodes_idx_set);
	int get_actual_node_num() { return n_nodes_involved; };
	int get_actual_edge_num() { return es.size(); };
	double density();
	double density_subgraph(std::vector<int>& subgraph_nodes_idx_set);
	void print(std::string out_file);
	void read_network_muleGraphFormat(const char* file, Edges& es);
	void read_network_muleGraphFormat(const char* file, std::vector<int>& selected_nodes_idx_set, Edges& es);
};

class Networks {
private:
	Network* nets;
	int n_net;
public:
	Networks() {};
	Networks(const char* nets_path, std::vector<std::string>& nets_IDs_list, const char* net_file_suffix,
		const char* file_type, std::vector<int>& selected_nodes_idx_set);
	~Networks() {std::cout << "deleting a set of networks" << std::endl; delete [] nets;}
	int size() { return n_net; }
	Network& operator [](int index);
	const Network& operator [](int index) const;
};

#endif