#include <cmath>
#include <set>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
#include <boost/algorithm/string.hpp>
#include <boost/dynamic_bitset.hpp>
#include "utils.h"
#include "Isomorphism.h"

using namespace std;
using namespace boost;

void print( vector<string>& v, const char* delim )
{
	for (size_t n = 0; n<v.size(); n++)
		cout << "\"" << v[n] << "\"" << delim;
	cout << endl;
}
void print( vector<int>& v, const char* delim )
{
	for (size_t n = 0; n<v.size(); n++)
		cout << v[n]+1 << delim;
	cout << endl;
}
void print( vector<ISOMORPHISM_CODING_NUMBER_TYPE>& v, const char* delim )
{
	for (size_t n = 0; n<v.size(); n++)
		cout << v[n] << delim;
	cout << endl;
}
void print( set<ISOMORPHISM_CODING_NUMBER_TYPE>& v, const char* delim )
{
	set<ISOMORPHISM_CODING_NUMBER_TYPE>::iterator iter;
	for (iter = v.begin(); iter != v.end(); ++iter)
		cout << *iter << delim;
	cout << endl;
}
void print( vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> >& v )
{
	vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> >::iterator iter;
	for (iter = v.begin(); iter != v.end(); ++iter)
		cout << (*iter).first << " <-> " << (*iter).second << endl;
}
void print( map<ISOMORPHISM_CODING_NUMBER_TYPE,int>& v )
{
	map<ISOMORPHISM_CODING_NUMBER_TYPE,int>::iterator iter;
	for (iter = v.begin(); iter != v.end(); ++iter)
		cout << (*iter).first << " -> " << (*iter).second << endl;
}
void print( ofstream& of, vector<int>& v )
{
	of << "[";
	int i;
	for (i=0; i<v.size()-1; i++)
		of << v[i]+1 << ", ";
	of << v[i]+1 << "]";
}
void print( ofstream& of, vector<int>& v, string prefix, string delimit, string suffix )
{
	of << prefix;
	int i;
	for (i=0; i<v.size()-1; i++)
		of << v[i]+1 << delimit;
	of << v[i]+1 << suffix;
}

void fin_open(ifstream& fin, const char* filename) {
    fin.open(filename);
    if (fin.fail()) {
        cerr << "Unable to open input file: \"" << filename << "\"" << endl;
        exit(1);
    }
}

void fin_open(ifstream& fin, string filename) {
	fin.open(filename.c_str());
    if (fin.fail()) {
        cerr << "Unable to open input file: " << filename << endl;
        exit(1);
    }
}

void rstrip(string &line, const char str_[]) {
    string str = str_;
    string::size_type pos = line.find(str, 0);
    if (pos != string::npos && pos == line.size()-str.size())
        line.erase(pos ,str.size());
}


void rstrip(string &line) {
    rstrip(line, "\n");
    rstrip(line, "\r");
    rstrip(line, "\n");
}

void split(string &line, const char *delim_, vector<string>& pieces) {
	pieces.clear();
    rstrip(line);
    string delim = delim_;
    string::size_type cur_pos = 0, next_delim=0;
    do {
        next_delim = line.find(delim, cur_pos);
        if (next_delim == string::npos)
            pieces.push_back(line.substr(cur_pos, line.size()-cur_pos));

        else {
            assert(cur_pos < line.size());
            assert(next_delim < line.size());
            pieces.push_back(line.substr(cur_pos, next_delim-cur_pos));
            cur_pos = next_delim + delim.size();
        }
    }while(next_delim != string::npos);
}

void split_int(string &line, char *delim_, vector<int>& split) {
	split.clear();
    rstrip(line);
    string delim = delim_;
    //cout << "splitting ints`" << line << endl;
    string::size_type cur_pos = 0, next_delim=0;
    do {
        next_delim = line.find(delim, cur_pos);
        if (next_delim == string::npos)
            split.push_back(atoi(line.substr(cur_pos, line.size()-cur_pos).c_str()));
        else {
            split.push_back(atoi(line.substr(cur_pos, next_delim-cur_pos).c_str()));
            cur_pos = next_delim + delim.size();
        }
    }while (next_delim != string::npos);
}


vector<string>  list_from_file_str(string file) {
    vector<string>  vec;
    ifstream fin;
    fin_open(fin, file);
    string str;
    while (fin >> str)
        vec.push_back(str);
	fin.close();
    return vec;
}

// a sample of chromosome2beads file (columns are delimited by TAB):
//C01.1   1   31
//C01.2   2   32
//C01.3   3   33
//C01.4   4   34
//C01.5   5   35
//C01.6   6   36
//C01.7   7   37
//C01.8   8   38
//C01.9   9   39
//C01.10  10  40
//C01.11  11  41
//C01.12  12  42
//C01.13  13  43
//C01.14  14  44
//C01.15  15  45
//C01.16  16  46
//
// chrom_region_ID bead_copy_A_index bead_copy_B_index
// bead index are 1-based.
//
// Output:
//   "chromregion2bead" should be 0-based indexes of chrom_regions and beads.
//
void read_chromregions2beads_file(string chrom_region_ids_file,
	vector<string>& chrom_regions_IDs, map<int,vector<int> >& chromregion2bead)
{
	chrom_regions_IDs.clear();
	chromregion2bead.clear();
	vector<string> lines;
	ifstream fin;
	fin.open(chrom_region_ids_file.c_str());
	if (fin.fail()){
        cerr << "Error: Unable to open " << chrom_region_ids_file << " in read_chromregions2beads_file()" << endl;
        exit(EXIT_FAILURE);
    }
	string buffer;
	getline(fin, buffer);
	int i = 0;
	while (!fin.eof()) {
		vector<string> items;
		split(buffer, "\t", items);
//		cout << "\"" << lines[i] << "\"" << endl;
//		print(items, " | ");
		if (items.size()!=3) {
			cerr << "Error: each line of chrom_regions to beads file should have 3 columns!" << endl;
			exit(EXIT_FAILURE);
		}
		chrom_regions_IDs.push_back( items[0] );
		int bead_A = atoi(items[1].c_str()) - 1; // bead index should be corrected to be 0-based, although it is 1-based in the file
		int bead_B = atoi(items[2].c_str()) - 1; // bead index should be corrected to be 0-based, although it is 1-based in the file
		vector<int> bead_pair;
		bead_pair.push_back(bead_A);
		bead_pair.push_back(bead_B);
		chromregion2bead[i] = bead_pair;
		getline(fin, buffer);
		i++;
	}
	fin.close();
}

void get_chromosomes_of_chromregions(vector<string>& chrom_regions_IDs, vector<int>& chrom_regions,
	vector<string>& results)
{
	results.clear();
	for (int i=0; i<chrom_regions.size(); i++) {
		int chrom_region_idx = chrom_regions[i];
		vector<string> items;
		split( chrom_regions_IDs[chrom_region_idx], ".", items );
		results.push_back( items[0] );
	}
}

void get_strings(vector<string>& all_strings, vector<int>& select_indexes,
	vector<string>& result_strings)
{
	result_strings.clear();
	int n = all_strings.size();
	for (int i=0; i<select_indexes.size(); i++)
		if (select_indexes[i]<n)
			result_strings.push_back(all_strings[select_indexes[i]]);
}

// Input:
//   nodes: a set of nodes indexes
//   node2subnodes: a map for node_index -> its_subnodes_indexes
//
// Output:
//   subnodes: a set of subnodes indexes
//
void get_subnodes_of_nodes(dynamic_bitset<>& nodes_bits_encoding, vector<int>& nodes,
	map<int,vector<int> >& node2subnodes, vector<int>& subnodes)
{
	subnodes.clear();
	for (int i=0; i<nodes_bits_encoding.size(); i++) {
		int node_idx = nodes[i];
		int rank_subnode_in_node = nodes_bits_encoding[i]; // only two values: 0 or 1.
		subnodes.push_back(node2subnodes[node_idx][rank_subnode_in_node]);
	}
}

bool descending_order_int(pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> i, pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> j)
{
	return (i.second > j.second);
}

bool descending_order_double(pair<ISOMORPHISM_CODING_NUMBER_TYPE,double> i, pair<ISOMORPHISM_CODING_NUMBER_TYPE,double> j)
{
	return (i.second > j.second);
}

// note that each set in "transaction" is already sorted in increasing order.
// actually we doesn't use "freq_thresh".
void get_freq_items(vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> >& transactions,
	int freq_thresh, vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> >& freqs)
{
	// get unique items which are stored in a set (in increasing order).
	set<ISOMORPHISM_CODING_NUMBER_TYPE> union_;
	for (int i=0; i<transactions.size(); i++) {
		set_union(transactions[i].begin(), transactions[i].end(), union_.begin(), union_.end(), inserter(union_, union_.end()));
	}
//	print(union_, ", "); // for debug
	// get frequency of each item (by building a map "item2freq")
	set<ISOMORPHISM_CODING_NUMBER_TYPE>::iterator iter;
	map<ISOMORPHISM_CODING_NUMBER_TYPE,int> item2freq;
	for (iter = union_.begin(); iter!=union_.end(); ++iter)
		item2freq[*iter] = 0;
	for (int i=0; i<transactions.size(); i++)
		for (iter = transactions[i].begin(); iter != transactions[i].end(); ++iter)
			item2freq[*iter]++;
	freqs.clear();
	for (iter = union_.begin(); iter!=union_.end(); ++iter)
		freqs.push_back(make_pair(*iter,item2freq[*iter]));
	sort(freqs.begin(), freqs.end(), descending_order_int);
//	print( item2freq ); // for debug
}

void get_occurrences(vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> >& transactions, ISOMORPHISM_CODING_NUMBER_TYPE item,
	vector<int>& occurrences)
{
	occurrences.clear();
	set<ISOMORPHISM_CODING_NUMBER_TYPE>::iterator iter;
	for (int i=0; i<transactions.size(); i++)
		for (iter = transactions[i].begin(); iter != transactions[i].end(); ++iter)
			if (*iter == item)
				occurrences.push_back(i);

}