#ifndef UTILS_H
#define UTILS_H

#include <vector>
#include <iostream>
#include <set>
#include <map>
#include <list>
#include <string>
#include <boost/dynamic_bitset.hpp>
#include "Isomorphism.h"

using std::vector;
using std::list;
using std::string;
using std::ifstream;
using std::istream;
using std::ofstream;
using std::ostream;
using std::map;
using namespace boost;

// print
void print( vector<string>& v, const char* delim );
void print( vector<int>& v, const char* delim );
void print( vector<ISOMORPHISM_CODING_NUMBER_TYPE>& v, const char* delim );
void print( set<ISOMORPHISM_CODING_NUMBER_TYPE>& v, const char* delim );
void print( vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> >& v );
void print( ofstream& of, vector<int>& v );
void print( ofstream& of, vector<int>& v, string prefix, string delimit, string suffix );

// strings
void rstrip(string &line, const char str_[]);
void rstrip(string &line);
void split(string &line, const char *delim_, vector<string>& pieces);
void split_int(string &line, char *delim_, vector<int>& split);
void get_strings(vector<string>& all_strings, vector<int>& select_indexes,
	vector<string>& result_strings);

// Read files
vector<string>  list_from_file_str(string file);
void read_module_from_file(char* filename, int which_module, vector<int>& nodes, vector<int>& networks);
void read_chromregions2beads_file(string chrom_region_ids_file,	vector<string>& chrom_regions_IDs,
	map<int,vector<int> >& chromregion2bead);
void get_chromosomes_of_chromregions(vector<string>& chrom_regions_IDs, vector<int>& chrom_regions,
	vector<string>& results);
void get_subnodes_of_nodes(dynamic_bitset<>& nodes_bits_encoding, vector<int>& nodes,
	map<int,vector<int> >& node2subnodes, vector<int>& subnodes);

// frequent items
bool descending_order_int(pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> i, pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> j);
bool descending_order_double(pair<ISOMORPHISM_CODING_NUMBER_TYPE,double> i, pair<ISOMORPHISM_CODING_NUMBER_TYPE,double> j);
void get_freq_items(vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> >& transactions,
	int freq_thresh, vector<pair<ISOMORPHISM_CODING_NUMBER_TYPE,int> >& freqs);
void get_occurrences(vector<set<ISOMORPHISM_CODING_NUMBER_TYPE> >& transactions, ISOMORPHISM_CODING_NUMBER_TYPE item,
	vector<int>& occurrences);


#endif